/**
 * @file
 * JavaScript functions for Routes administration functionality.
 */

const { __, _x, _n, _nx } = wp.i18n;

/**
 * Looks up elevation for one point.
 *
 * @param lat_field
 *   DOM input element to read latitude from.
 * @param long_field
 *   DOM input element to read longitude from.
 * @param elev_field
 *   DOM input element to put elevation for this latitude/longitude in.
 *   Global variable routes_use_metric governs whether metric units are used
 *   or not.
 */
function routes_lookup_elev(lat_field, long_field, elev_field) {
  if (!lat_field || !lat_field.value.length ||
      !long_field || !long_field.value.length) {
      alert(__('Latitude and longitude not set', 'routes'));
    return true;
  }

  var data = {
    action: 'routes_elevation_lookup',
    points: [{
      latitude: parseFloat(lat_field.value),
      longitude: parseFloat(long_field.value)
    }]
  };
  jQuery.post(ajaxurl, data, function (results) {
    var output = JSON.parse(results);
    var result = output.results[0];
    var elev = result.elevation;
    if (elev < 0) {
      alert(__('Error looking up elevation. Debug information:', 'routes') + "\n" + results );
      return true;
    }
    if (routes_use_metric) {
      elev *= 3.280833;
    }
    elev_field.value = elev;
  });
  return true;
}

/**
 * Looks up elevation for 100 points, starting at routes_all_elev_start index.
 */
function routes_lookup_100_elev() {
  // Go through the points and form an array.
  var locations = [];
  var routes_form = routes_form_ptr();
  var start = parseFloat(routes_form['routes_all_elev_start'].value);
  if (start < 1) {
    start = 1;
  }
  for (var i = start; i <= start + 100 && i <= routesFormNumLines; i++) {
    if (routes_form[ "routes_delete" + i ].checked) {
      continue;
    }
    if (routes_form[ "routes_order" + i ].value.length < 1 ||
        routes_form[ "routes_lat" + i ].value.length < 1 ||
        routes_form[ "routes_long" + i ].value.length < 1) {
      continue;
    }

    locations[locations.length] = {
      latitude: parseFloat(routes_form[ "routes_lat" + i ].value),
      longitude: parseFloat(routes_form[ "routes_long" + i ].value)
    };
  }

  var data = {
      action: 'routes_elevation_lookup',
      points: locations
  };
  jQuery.post(ajaxurl, data, function (results) {
    var output = JSON.parse(results);
    var form_i = parseFloat(routes_form['routes_all_elev_start'].value);
    if (form_i < 1) {
        form_i = 1;
    }
    for(var results_i = 0; results_i < output.results.length;) {
      if (routes_form[ "routes_delete" + form_i ].checked) {
        form_i++;
        continue;
      }
      if (routes_form[ "routes_order" + form_i ].value.length < 1 ||
        routes_form[ "routes_lat" + form_i ].value.length < 1 ||
        routes_form[ "routes_long" + form_i ].value.length < 1) {
          form_i++;
          continue;
      }
      result = output.results[results_i];
      results_i++;
      var elev = result.elevation;
      if (elev < 0) {
          alert(__('Error looking up elevation. Debug information:', 'routes') + "\n" + JSON.stringify(result));
        return true;
      }
      if (routes_use_metric) {
        elev *= 3.280833;
      }
      routes_form["routes_elev" + form_i].value = elev;
      form_i++;
    }
    routes_refresh_map();
    routes_form['routes_all_elev_start'].value = form_i;

  });

  return true;
}

/**
 * Finds the index of the point nearest to the given latitude/longitude.
 */
function routes_find_index(lat, long) {
  var want_blank = (lat < -200 && long < -200);
  var tol = 0.0005;
  var routes_form = routes_form_ptr();
  var best_index = 0;
  var best_tol = tol + 100;

  for (i = 1; i <= routesFormNumLines; i++) {
    var thisLong = parseFloat(routes_form[ "routes_long" + i ].value);
    var thisLat = parseFloat(routes_form[ "routes_lat" + i ].value);
    var thisLen = routes_form[ "routes_order" + i ].value.length +
      routes_form[ "routes_city" + i ].value.length +
      routes_form[ "routes_state" + i ].value.length +
      routes_form[ "routes_lat" + i ].value.length +
      routes_form[ "routes_long" + i ].value.length;

    // are we looking for a blank line?
    if (want_blank && thisLen < 5) {
      return i;
    }
    // are we looking for a match?
    var this_tol = Math.abs(thisLong - long) + Math.abs(thisLat - lat);
    if (this_tol < best_tol) {
      best_tol = this_tol;
      best_index = i;
    }
  }

  if (best_tol < tol && !want_blank) {
    return best_index;
  }

  return -1;
}

/**
 * Finds the POI nearest the given latitude/longitude.
 */
function routes_find_poi_index(lat, long) {
  var want_blank = (lat < -200 && long < -200);
  var tol = 0.0005;
  var routes_form = routes_form_ptr();
  var best_index = 0;
  var best_tol = tol + 100;

  for (i = 1; i <= routesFormNumPOILines; i++) {
    var thisLong = parseFloat(routes_form[ "routes_poi_long" + i ].value);
    var thisLat = parseFloat(routes_form[ "routes_poi_lat" + i ].value);
    var thisLen = routes_form[ "routes_poi_lat" + i ].value.length +
      routes_form[ "routes_poi_long" + i ].value.length;

    // are we looking for a blank line?
    if (want_blank && thisLen < 5) {
          return i;
    }
    // are we looking for a match?
    var this_tol = Math.abs(thisLong - long) + Math.abs(thisLat - lat);
    if (this_tol < best_tol) {
      best_tol = this_tol;
      best_index = i;
    }
  }

  if (best_tol < tol && !want_blank) {
    return best_index;
  }

  return -1;
}

/**
 * Adds a listener to a marker to delete the corresponding waypoint.
 */
function routes_setup_remove_listener(lat, long, marker) {
  marker.on('click', function() {
    // find point and mark as deleted
    var indx = routes_find_index(lat, long);
    var routes_form = routes_form_ptr();
    if (indx > 0) {
      routes_form[ "routes_delete" + indx ].checked = 1;
      routes_refresh_map();
      return true;
    }

    alert(__("Could not find waypoint in table. Click Refresh Map or Save and Continue Editing.", 'routes'));
    return false;
  });
}

/**
 * Adds a listener to a marker to delete the corresponding POI.
 */
function routes_setup_poi_remove_listener(lat, long, marker) {
  marker.on('click', function() {
    // find POI and mark as deleted
    var indx = routes_find_poi_index(lat, long);
    var routes_form = routes_form_ptr();
    if (indx > 0) {
      routes_form[ "routes_poi_delete" + indx ].checked = 1;
      routes_refresh_map();
      return true;
    }

    alert(__("Could not find point of interest in table. Please click Refresh Map or Save and Continue Editing.", 'routes'));
    return false;
  });
}

/**
 * Adds more waypoint lines to the edit form.
 */
function routes_add_more_lines() {
  var txt = '';
  var oldnum = routesFormNumLines;

  // Make sure the last section is shown.
  jQuery("." + routesFormLastClass).removeClass('routes_group_hidden');

  for (i = 0; i < routes_blank_lines; i++) {
    routesFormNumLines++;
    txt += '<tr class="' + routesFormLastClass + '">';
    txt += '<td><input name="routes_order' + routesFormNumLines + '" type="text" size="3"/></td>';
    txt += '<td><input name="routes_city' + routesFormNumLines + '" type="text" size="15"/></td>';
    txt += '<td><input name="routes_state' + routesFormNumLines + '" type="text" size="7"/></td>';
    txt += '<td><input name="routes_lat' + routesFormNumLines + '" type="text" size="6"/></td>';
    txt += '<td><input name="routes_long' + routesFormNumLines + '" type="text" size="6"/></td>';

    // checkbox for delete it
    txt += '<td><input class="delcheck" type="checkbox" name="routes_delete' +
      routesFormNumLines + '"/></td>' + "\n";

    txt += "</tr>\n";

    // second row for metadata

    txt += '<tr class="' + routesFormLastClass + ' routes_extended';
    if (routesInfoHidden) {
      txt += ' routes_extended_hidden';
    }

    txt += '"' + ">\n";
    txt += '<td colspan="5">' + "\n";
    txt += __('Tag:', 'routes');
    txt += '<input name="routes_tag' + routesFormNumLines + '" type="text" size="35" /> ';
    txt += __('Elevation:', 'routes');

    txt += '<input type="button"' +
      ' name="routes_elev_btn' + routesFormNumLines + '" id="routes_elev_btn' +
      routesFormNumLines + '" value="' +
      __('Lookup', 'routes') +
      '" onclick="routes_lookup_elev(' +
      "this.form.routes_lat" + routesFormNumLines + "," +
      "this.form.routes_long" + routesFormNumLines + "," +
      "this.form.routes_elev" + routesFormNumLines + ");return true;" + '"' +
      "/>";

    txt += '<input name="routes_elev' + routesFormNumLines +
      '" id="routes_elev' + routesFormNumLines +
      '" type="text" size="6" /> ';
    txt += "</td></tr>\n";
  }

  jQuery('#routes_elev_btn' + oldnum).parent().parent().after(txt);
}

/**
 * Adds more POI lines to the edit form.
 */
function routes_add_more_poi_lines() {
  var txt = '';
  var oldnum = routesFormNumPOILines;

  // Make sure the last section is shown.
  jQuery("." + routesFormLastPOIClass).removeClass('routes_group_hidden');

  for (i = 0; i < routes_blank_lines; i++) {
    routesFormNumPOILines++;
    txt += '<tr class="' + routesFormLastPOIClass + '">';
    txt += '<td><input name="routes_poi_lat' + routesFormNumPOILines + '" type="text" size="6"/></td>';
    txt += '<td><input name="routes_poi_long' + routesFormNumPOILines + '" type="text" size="6"/></td>';
    txt += '<td>' + routesIconSelect.replace('/replace_name_here/', 'routes_poi_icon' + routesFormNumPOILines) + '</td>';

    // checkbox for delete it
    txt += '<td><input class="delcheck" type="checkbox" name="routes_poi_delete' +
      routesFormNumPOILines + '"/></td>' + "\n";

    txt += "</tr>\n";

    txt += '<tr class="' + routesFormLastPOIClass + '"' + ">\n";
    txt += '<td colspan="3">' + "\n";
    txt += __('Text:', 'routes');
    txt += '<input name="routes_poi_tag' + routesFormNumPOILines + '" id="routes_poi_tag' + routesFormNumPOILines + '" type="text" size="50" /> ';
    txt += "</td></tr>\n";
  }

  jQuery('#routes_poi_tag' + oldnum).parent().parent().after(txt);
}

/**
 * Adds a waypoint, after clicking on the given latitude and longitude.
 */
function routes_add_waypoint(lat, long) {
  var indx = routes_find_index(-999, -999);
  if (indx <= 0) {
    routes_add_more_lines();
    indx = routes_find_index(-999, -999);
  }

  var order = 10;
  var routes_form = routes_form_ptr();
  if (indx > 0) {
    routes_form[ "routes_lat" + indx ].value = lat;
    routes_form[ "routes_long" + indx ].value = long;
    if (indx > 1) {
      order = parseFloat(routes_form[ "routes_order" + (indx - 1) ].value) + 10;
      routes_form[ "routes_city" + indx ].value =
        routes_form[ "routes_city" + (indx - 1) ].value;
      routes_form[ "routes_state" + indx ].value =
        routes_form[ "routes_state" + (indx - 1) ].value;
    }
    routes_form[ "routes_order" + indx ].value = order;

    if (routes_auto_elev) {
      routes_lookup_elev(routes_form[ "routes_lat" + indx ],
                          routes_form[ "routes_long" + indx ],
                          routes_form[ "routes_elev" + indx ]);
    }

    return true;
  }

  alert(__("Could not find blank line in table. Please click Save and Continue Editing to get more lines.", 'routes'));
  return false;
}

/**
 * Adds a POI, after clicking on the given latitude and longitude.
 */
function routes_add_poi_point(lat, long) {
  var indx = routes_find_poi_index(-999, -999);
  if (indx <= 0) {
    routes_add_more_poi_lines();
    indx = routes_find_poi_index(-999, -999);
  }

  var order = 10;
  var routes_form = routes_form_ptr();
  if (indx > 0) {
    routes_form[ "routes_poi_lat" + indx ].value = lat;
    routes_form[ "routes_poi_long" + indx ].value = long;
    return true;
  }

  alert(__("Could not find blank line in table. Please click Save and Continue Editing to get more lines.", 'routes'));
  return false;
}

/**
 * Adds a POI point or waypoint at the given latitude/longitude.
 */
function routes_add_point(lat, long) {
  var routes_form = routes_form_ptr();
  if (routes_form['routes_click_action'][0]. checked) {
    routes_add_waypoint(lat, long);
  }
  else {
    routes_add_poi_point(lat, long);
  }
}

/**
 * Refreshes the map.
 */
function routes_refresh_map() {
  var routes_form = routes_form_ptr();

  // First, clear all route markers and lines.
  for (var i = 0; i < routesNumOverlays; i++) {
    routesOverlays[i].remove();
    routesOverlays[i] = null;
  }

  // Now add markers for waypoints.
  var rpts = [];

  routesNumOverlays = 0;
  var show_every = 1;
  if (routesFormNumLines > 60) {
    show_every = Math.ceil(routesFormNumLines/50);
  }

  var shown = 1;
  for (i = 1; i <= routesFormNumLines; i++) {
    if (routes_form[ "routes_delete" + i ].checked) {
      continue;
    }
    if (routes_form[ "routes_order" + i ].value.length < 1 ||
        routes_form[ "routes_lat" + i ].value.length < 1 ||
        routes_form[ "routes_long" + i ].value.length < 1) {
      continue;
    }

    var thisLong = parseFloat(routes_form[ "routes_long" + i ].value);
    var thisLat = parseFloat(routes_form[ "routes_lat" + i ].value);
    var thisOrder = routes_form[ "routes_order" + i ].value;
    var thisTitle = thisOrder + ": " + thisLat + ", " + thisLong;

    // create a point and add to arrays
    var pt = L.latLng(thisLat, thisLong);
    rpts[ rpts.length] = pt;
    if (shown < show_every) {
      shown++;
    }
    else {
      var mrk = L.marker(pt, { title: thisTitle });
      mrk.addTo(routesmap);
      routes_setup_remove_listener(thisLat, thisLong, mrk);
      routesOverlays[routesNumOverlays] = mrk;
      routesNumOverlays++;
      shown = 1;
    }
  }

  // put the route on screen

  if (rpts.length > 1) {
    var routesline = L.polyline(rpts, {color: "#0000ff", weight: 3, opacity: .5 });
    routesline.addTo(routesmap);
    routesOverlays[routesNumOverlays] = routesline;
    routesNumOverlays++;
  }

  // Now do Points of Interest
  for (i = 1; i <= routesFormNumPOILines; i++) {
    if (routes_form[ "routes_poi_delete" + i ].checked) {
      continue;
    }
    if (routes_form[ "routes_poi_lat" + i ].value.length < 1 ||
        routes_form[ "routes_poi_long" + i ].value.length < 1) {
      continue;
    }

    var thisLong = parseFloat(routes_form[ "routes_poi_long" + i ].value);
    var thisLat = parseFloat(routes_form[ "routes_poi_lat" + i ].value);
    var iconnum = parseInt(routes_form[ "routes_poi_icon" + i ].value);

    var pt = L.latLng(thisLat, thisLong);
    var mrk = L.marker(pt, {icon: routes_icons[iconnum] });
    mrk.addTo(routesmap);
    routes_setup_poi_remove_listener(thisLat, thisLong, mrk);
    routesOverlays[routesNumOverlays] = mrk;
    routesNumOverlays++;
  }

  return true;
}

/**
 * Toggles whether to show/hide extended information in editing table.
 */
function routes_showall() {
  jQuery('tr.routes_extended').toggleClass('routes_extended_hidden');
  routesInfoHidden = !routesInfoHidden;
}

/**
 * Marks all points for deletion.
 */
function routes_mark_all_del() {
  // marks all waypoints for delete
  jQuery('input.delcheck').attr('checked', 1);
}

/**
 * Toogles whether one group of waypoints or POI points is shown/hidden.
 */
function routes_group_showhide(grpid) {
   jQuery('tr.' + grpid).toggleClass('routes_group_hidden');
}
