<?php
/*
Plugin Name: Routes
Plugin URI: https://poplarware.com/downloads/routes
Description: Adds route-based geographical information to your blog.
Version: 3.0
Author: Jennifer Hodgdon, Poplar ProductivityWare
Author URI: https://poplarware.com
Text Domain: routes
License: GNU General Public License v2 or later

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.
*/

// Functions for use in themes --------------------------

/**
 * Puts a map on the screen for a single post/page, if it has waypoints.
 *
 * Example usage in the single.php theme file:
 * @code
 * routes_map(get_the_ID(), 600, 600, "Routes shown are approximate.");
 * @endcode
 *
 * @param int $id
 *   ID of the post or page to put up the map for.
 * @param int $width
 *   (optional) Width of map in pixels (default 500).
 * @param int $height
 *   (optional) Height of map in pixels (default 400).
 * @param string $printnote
 *   (optional) Note to print above the map.
 */
function routes_map($id, $width=500, $height=400, $printnote='') {
  routes_map_print(routes_read_waypoints($id), routes_read_poi($id),
    $width, $height, "post", true, $printnote);
}

/**
 * Puts a map on the screen for an archive page, if there are waypoints.
 *
 * Example usage in the category.php theme file (can also be used for search):
 * @code
 * routes_map_multiple(600, 600, "Click on a marker or line to see more info");
 * @endcode
 *
 * @param int $width
 *   (optional) Width of map in pixels (default 500).
 * @param int $height
 *   (optional) Height of map in pixels (default 400).
 * @param string $printnote
 *   (optional) Note to print above the map.
 */
function routes_map_multiple($width=500, $height=400, $printnote='') {
  global $wp_query;

  if (!$wp_query->post_count) {
    return;
  }

  $ids = array();
  foreach ($wp_query->posts as $post) {
    $ids[] = $post->ID;
  }

  routes_map_print(routes_read_waypoints($ids), array(), $width, $height,
    "multiple", true, $printnote);
}

/**
 * Prints the elevation profile image for a single post/page, if it exists.
 *
 * Example for single.php theme file:
 * @code
 * routes_elev_profile(get_the_ID(), "Feet vs. miles traveled (approx).");
 * @endcode
 *
 * @param int $id
 *   Post or page ID.
 * @param string $printnote
 *   (optional) Note to print above the map.
 */
function routes_elev_profile($id, $printnote='') {
  $fname = routes_elev_graph_filename($id);

  $uploads = routes_graph_dir();
  if ($uploads[ 'error' ]) {
    return;
  }

  $path = $uploads[ 'path' ] . "/" . $fname;
  if (!file_exists($path)) {
    return;
  }

  if (strlen($printnote)) {
    echo '<p class="elevlegend">' . "$printnote</p>\n";
  }

  echo '<img src="' . $uploads[ 'url' ] . "/" . $fname . '"' .
    ' alt="' . __("Elevation Profile", 'routes') . '" />' . "\n";
}

/**
 * Prints a link to the GPX download for the current post, if there is one.
 *
 * Put this function inside the loop.
 *
 * @param string $linktext
 *   (optional) Link text.
 * @param int $id
 *   (optional) Override of the post ID.
 */
function routes_gpx_link($linktext = 'Download GPX', $id = NULL) {
  if (!isset($id)) {
    $id = get_the_ID();
  }
  $id = intval($id);

  if (routes_post_has_route($id)) {
    echo '<a href="' . home_url() . '?routes_gpx=' . $id . '">' . $linktext . '</a>';
  }
}

/**
 * Prints a search form, including city and state/country if desired.
 *
 * @param string $genLabel
 *   (optional) Label for general search field; set to '' to omit this field.
 * @param string $cityLabel
 *   (optional) Label for city search field; set to '' to omit this field.
 * @param string $stateLabel
 *   (optional) Label for state search field; set to '' to omit this field.
 * @param string $buttonText
 *   (optional) Text for button.
 */
function routes_search_form($genLabel = 'Keyword', $cityLabel = 'City',
  $stateLabel = 'State/Country', $buttonText = 'Search') {

  echo '<div><form method="get" id="searchform" action="';
  bloginfo('home');
  echo '">' . "\n";

  if (strlen($genLabel) > 0) {
    echo '<input type="text" name="s" id="s" /> ' .
      '<label for="s">' . $genLabel . "</label><br />\n";
  }

  if (strlen($cityLabel) > 0) {
    echo '<input type="text" name="routes_city" id="routes_city" /> ' .
      '<label for="routes_city">' . $cityLabel . "</label><br />\n";
  }

  if (strlen($stateLabel) > 0) {
    echo '<input type="text" name="routes_state" id="routes_state" /> ' .
      '<label for="routes_state">' . $stateLabel . "</label><br />\n";
  }

  echo '<input type="submit" id="searchsubmit" value="' . $buttonText .
    '" /></form></div>' . "\n";
}

/**
 * Prints a list of states that exist in the route points database, as links.
 *
 * @param string $sep
 *   Separator to use between state links.
 * @param int $postID
 *   (optional) If set, limit to states on this post/page ID.
 */
function routes_list_states($sep = ", ", $postID = 0) {
  global $routesWaypointTable, $wpdb;

  // find list of states in DB for published posts/pages
  $where = "(post_type = 'post' OR post_type = 'page') AND post_status = 'publish'";
  $postID = intval($postID);
  if ($postID) {
    $where = "ID = $postID";
  }

  $qry = "SELECT state_country FROM $routesWaypointTable JOIN " . $wpdb->posts . " ON $routesWaypointTable.post_id = " . $wpdb->posts . ".ID WHERE $where GROUP BY state_country ORDER BY state_country";
  $states = $wpdb->get_results($qry);

  // make list of links

  $before = '<a href="' . get_bloginfo('home') . '?routes_state=';
  $mid = '">';
  $after = "</a>";
  $cur_sep = "";
  foreach ($states as $row) {
    $state = $row->state_country;
    if (strlen($state)) {
      echo $cur_sep . $before . rawurlencode($state) . $mid . $state . $after;

      // after the first time, we need separator
      $cur_sep = $sep;
    }
  }
}

/**
 * Checks to see if there is a route for the current post.
 *
 * @param int $id
 *   (optional) Post ID (if omitted, uses the current post ID).
 *
 * @return bool
 *   TRUE if there is a route for the current post; FALSE otherwise.
 */
function routes_post_has_route($id = NULL) {
  global $wpdb;
  global $routesWaypointTable;

  if (!isset($id)) {
    $id = get_the_ID();
  }
  $id = intval($id);

  // See if there is a point.
  $qry = "SELECT waypoint_id FROM $routesWaypointTable WHERE post_id=$id LIMIT 1";
  $out = $wpdb->get_results($qry);
  if (count($out)) {
    return TRUE;
  }
  return FALSE;
}

/**
 * Prints out a legend of icons.
 */
function routes_icon_legend() {
  global $wpdb;
  global $routesIconTable;

  echo '<table class="routes-icon-legend">';
  $qry = "SELECT * FROM $routesIconTable";
  $icons = $wpdb->get_results($qry);
  if (isset($icons) && count($icons)) {
    foreach ($icons as $icon) {
      if ($icon->legend_name) {
        echo '<tr><td><img src="' . $icon->icon_image . '" alt="" /></td><td>' . $icon->legend_name . '</td></tr>';
      }
    }
  }
  echo '</table>';
}

/**
 * Checks to see if this is a keyword, city, or state search.
 *
 * @return bool
 *   TRUE if this is a keyword, city, or state search page; FALSE if not.
 */
function routes_is_search() {
  global $wp_query;

  $q = $wp_query->query_vars;

  return (!empty($q['s']) ||
    !empty($q['routes_city']) ||
    !empty($q['routes_state'])
  );
}

// Global Variables and defines -----------------------------------

if (!defined('WP_CONTENT_DIR')) {
  define('WP_CONTENT_DIR', ABSPATH . 'wp-content');
}
if (!defined('WP_CONTENT_URL')) {
  define('WP_CONTENT_URL', get_option('siteurl') . '/wp-content');
}

// Routes options.
$routesScale = 0;
$routesIsSetUp = 0;
$routesJSForFooter = '';
$routesElevWidth = 0;
$routesElevHeight = 0;
$routesUseMetric = 0;
$routesMakeElev = 0;
$routesLookupElev = 0;
$routesLeafletElev = 0;
$routesElevMarkers = 0;
$routesElevLineColor = array(0, 0, 0);
$routesElevBgColor = array(255, 255, 255);
$routesElevFillColor = array(100, 100, 255);
$routesElevMarkerColor = array(255, 100, 100);
$routesEditHeaders = 15;

// Global declarations needed for plugin activation for some reason.
global $wpdb;
global $routesDBVersion;
global $routesWaypointTable;
global $routesPOITable;
global $routesIconTable;
global $routesPostIconTable;
global $routesPostLineStyleTable;
global $routesPostMapTypeTable;

$routesDBVersion = "4";
$routesWaypointTable = $wpdb->prefix . "routes_waypoint";
$routesPOITable = $wpdb->prefix . "routes_poi";
$routesIconTable = $wpdb->prefix . "routes_icon";
$routesPostIconTable = $wpdb->prefix . "routes_post_icon";
$routesPostLineStyleTable = $wpdb->prefix . "routes_post_line";
$routesPostMapTypeTable = $wpdb->prefix . "routes_post_maptype";

$routesFileName = plugin_basename(__FILE__);
$routesVerifyKey = "routes_on_edit_screen";
$routesBlankLines = 20;
//$routesGPXMimeType = 'application/gpx+xml';
$routesGPXMimeType = 'application/xml';

// Filters and actions ------------------------

add_action('activate_' . $routesFileName, 'routes_call_install_tables');
add_action('save_post', 'routes_call_post_update'); // this does pages too
add_action('delete_post', 'routes_call_post_delete'); // this does pages too
add_action('admin_menu', 'routes_call_admin_screen');
add_action('admin_ns', 'routes_admin_ns');
add_action('admin_xml_ns', 'routes_admin_ns');
add_action('wp_head', 'routes_js_header');
add_action('init', 'routes_init');
add_action('admin_print_styles', 'routes_css_admin_header');
add_action('admin_print_scripts', 'routes_call_js_admin_header');
add_action('admin_footer', 'routes_call_js_admin_footer', 6);
add_action('wp_footer', 'routes_do_map_js', 1000);
add_filter('posts_where', 'routes_search_where');
add_filter('posts_join', 'routes_search_join');
add_filter('posts_groupby', 'routes_search_groupby');
add_action('pre_get_posts', 'routes_search_fixquery');
add_filter('query_vars', 'routes_search_queryvars');
add_action('parse_query', 'routes_set_ishome');
add_filter('upload_mimes', 'routes_add_mimes');
add_filter('the_content', 'routes_map_in_post', 1);
add_action('after_plugin_row', 'routes_call_plugin_row', 10, 2);
add_action('wp_ajax_routes_elevation_lookup', 'routes_call_ajax_elevation_lookup');

// Admin Menu and Screen Functions --------------------------
//
// These functions are mostly located in an include file, so the code here just
// includes it and calls them. Except for some that are too short to bother
// with, which are documented.

function routes_call_plugin_row($file, $plugin_data) {
  include_once plugin_dir_path(__FILE__) . 'routes.admin.php';
  routes_plugin_row($file, $plugin_data);
}

function routes_call_install_tables() {
  include_once plugin_dir_path(__FILE__) . 'routes.admin.php';
  routes_install_tables();
}

function routes_call_admin_screen() {
  include_once plugin_dir_path(__FILE__) . 'routes.admin.php';
  routes_admin_screen();
}

function routes_call_post_update($id) {
  include_once plugin_dir_path(__FILE__) . 'routes.admin.php';
  routes_post_update($id);
}

function routes_call_post_delete($id) {
  include_once plugin_dir_path(__FILE__) . 'routes.admin.php';
  routes_post_delete($id);
}

function routes_call_js_admin_header() {
  include_once plugin_dir_path(__FILE__) . 'routes.admin.php';
  routes_js_admin_header();
}

function routes_call_js_admin_footer() {
  include_once plugin_dir_path(__FILE__) . 'routes.admin.php';
  routes_js_admin_footer();
}

function routes_call_ajax_elevation_lookup() {
  include_once plugin_dir_path(__FILE__) . 'routes.admin.php';
  routes_ajax_elevation_lookup();
}

/**
 * Prints the XML name space for IE for the admin screens.
 */
function routes_admin_ns() {
  echo 'xmlns:v="urn:schemas-microsoft-com:vml"';
}

/**
 * Adds some CSS to the admin screen.
 */
function routes_css_admin_header() {
  print '<style type="text/css">' . "\n" .
    'tr.routes_group_hidden, tr.routes_extended_hidden {display: none;}' . "\n" .
    '</style>' . "\n";
}

// Internal functions. ------------------------------------

/**
 * Returns an array of icons to use.
 *
 * @param int $id
 *   Post ID to get icons for, or -1 for defaults.
 *
 * @return array
 *   Icons to use for this post, or defaults if none have been defined.
 *
 * @internal
 */
function routes_get_post_icons($id = -1) {
  global $wpdb;
  global $routesPostIconTable;
  static $defaults = NULL;

  if (!isset($defaults) && $id != -1) {
    $defaults = routes_get_post_icons(-1);
  }
  // Get defaults, unless defaults were requested.
  if ($id != -1) {
    $ret = $defaults;
  }
  else {
    $ret = array();
  }

  // Sanitize.
  $id = intval($id);

  $qry = "SELECT icon_type, icon_id FROM $routesPostIconTable WHERE post_id = $id";
  $icons = $wpdb->get_results($qry);
  foreach ($icons as $icon) {
    $index = (int) routes_fix_num($icon->icon_type, 1, 3);
    $ret[$index - 1] = $icon->icon_id;
  }

  return $ret;
}

/**
 * Returns the line style to use for the given post.
 *
 * @param int $id
 *   Post ID to get line style for, or -1 for defaults.
 *
 * @return array
 *   Line style to use for this post, or defaults if none has been defined.
 *
 * @internal
 */
function routes_get_post_line_style($id = -1) {
  global $wpdb;
  global $routesPostLineStyleTable;
  static $defaults = NULL;

  if (!isset($defaults) && $id != -1) {
    $defaults = routes_get_post_line_style(-1);
  }

  // Sanitize.
  $id = intval($id);

  $qry = "SELECT t.color, t.opacity, t.weight FROM $routesPostLineStyleTable t WHERE post_id = $id";
  $styles = $wpdb->get_row($qry, ARRAY_N);
  if ($styles) {
    return $styles;
  }

  return $defaults;
}

/**
 * Prints out the Search widget.
 */
function routes_widget_search($args) {
  global $routesWidgetSearchTitle, $routesWidgetSearchKeywordLabel,
    $routesWidgetSearchCityLabel,
    $routesWidgetSearchStateLabel,
    $routesWidgetSearchSearchLabel;

  extract($args);

  echo $before_widget;
  echo $before_title;
  echo $routesWidgetSearchTitle;
  echo $after_title;

  routes_search_form($routesWidgetSearchKeywordLabel,
                      $routesWidgetSearchCityLabel,
                      $routesWidgetSearchStateLabel,
                      $routesWidgetSearchSearchLabel);
  echo $after_widget;
}

/**
 * Prints and/or saves the options screen for the search widget.
 */
function routes_widget_search_control() {
  global $routesWidgetSearchTitle, $routesWidgetSearchKeywordLabel,
    $routesWidgetSearchCityLabel,
    $routesWidgetSearchStateLabel,
    $routesWidgetSearchSearchLabel;

  if ($_POST["routes_search_options_submit"]) {
    $routesWidgetSearchTitle =
      $_POST["routes_search_title"];
    $routesWidgetSearchKeywordLabel =
      $_POST["routes_search_keyword_label"];
    $routesWidgetSearchCityLabel =
      $_POST["routes_search_city_label"];
    $routesWidgetSearchStateLabel =
      $_POST["routes_search_state_label"];
    $routesWidgetSearchSearchLabel =
      $_POST["routes_search_search_label"];

    update_option('routes_widget_search_title', $routesWidgetSearchTitle);
    update_option('routes_widget_search_keyword_label', $routesWidgetSearchKeywordLabel);
    update_option('routes_widget_search_city_label', $routesWidgetSearchCityLabel);
    update_option('routes_widget_search_state_label', $routesWidgetSearchStateLabel);
    update_option('routes_widget_search_search_label', $routesWidgetSearchSearchLabel);

  }

?>
			<p><label for="routes_search_title"><?php _e('Title for Search Form', 'routes'); ?></label> <input name="routes_search_title" id="routes_search_title" type="text" width="20" value="<?php echo $routesWidgetSearchTitle; ?>" /></p>
			<p><label for="routes_search_keyword_label"><?php _e('Label for Keyword field', 'routes'); ?></label> <input name="routes_search_keyword_label" id="routes_search_keyword_label" type="text" width="20" value="<?php echo $routesWidgetSearchKeywordLabel; ?>" /></p>
			<p><label for="routes_search_city_label"><?php _e('Label for City field', 'routes'); ?></label> <input name="routes_search_city_label" id="routes_search_city_label" type="text" width="20" value="<?php echo $routesWidgetSearchCityLabel; ?>" /></p>
			<p><label for="routes_search_state_label"><?php _e('Label for State/Country field', 'routes'); ?></label> <input name="routes_search_state_label" id="routes_search_state_label" type="text" width="20" value="<?php echo $routesWidgetSearchStateLabel; ?>" /></p>
			<p><label for="routes_search_search_label"><?php _e('Label for Search button', 'routes'); ?></label> <input name="routes_search_search_label" id="routes_search_search_label" type="text" width="20" value="<?php echo $routesWidgetSearchSearchLabel; ?>" /></p>

			<input type="hidden" name="routes_search_options_submit" value="1" />
<?php
}

/**
 * Prints the widget for the state list.
 */
function routes_widget_state_list($args) {
  global $routesWidgetStatesTitle, $routesWidgetStatesSep;

  extract($args);

  echo $before_widget;
  echo $before_title;
  echo $routesWidgetStatesTitle;
  echo $after_title;

  routes_list_states($routesWidgetStatesSep);

  echo $after_widget;
}

/**
 * Prints and/or saves the options screen for the state list.
 */
function routes_widget_state_list_control() {
  global $routesWidgetStatesTitle, $routesWidgetStatesSep;

  if ($_POST["routes_states_options_submit"]) {
    $routesWidgetStatesTitle =
      $_POST["routes_states_title"];
    $routesWidgetStatesSep =
      $_POST["routes_states_sep"];

    update_option('routes_widget_states_title', $routesWidgetStatesTitle);
    update_option('routes_widget_states_sep', $routesWidgetStatesSep);
  }

?>
			<p><label for="routes_states_title"><?php _e('Title for States List', 'routes'); ?></label> <input name="routes_states_title" type="text" id="routes_states_title" width="20" value="<?php echo $routesWidgetStatesTitle; ?>" /></p>

			<p><label for="routes_states_sep"><?php _e('Separator for States List', 'routes'); ?></label> <input name="routes_states_sep" type="text" id="routes_states_sep" width="5" value="<?php echo $routesWidgetStatesSep; ?>" /></p>

			<input type="hidden" name="routes_states_options_submit" value="1" />
<?php
}

/**
 * Sets up the Routes plugin.
 *
 * Initializes options, registers the sidebar widgets, and intercepts
 * the GPX download request if present.
 */
function routes_init() {

  routes_setup();

  $post_id = intval($_GET['routes_gpx']);
  if ($post_id) {
    routes_send_gpx($post_id);
  }

  wp_register_sidebar_widget('routes-search',
    __('Routes Search', 'routes'),
    'routes_widget_search');

  wp_register_sidebar_widget('routes-state-list',
    __('Routes State List', 'routes'),
    'routes_widget_state_list');

  wp_register_widget_control('routes-search',
    __('Routes Search', 'routes'),
    'routes_widget_search_control');

  wp_register_widget_control('routes-state-list',
    __('Routes State List', 'routes'),
    'routes_widget_state_list_control');

  wp_register_script('routes-admin-js',
    plugin_dir_url(__FILE__) . 'routes-admin.js',
    array('jquery', 'wp-i18n'));

  wp_register_script('routes-page-js',
    plugin_dir_url(__FILE__) . 'routes-page.js',
    array('jquery'));
}

/**
 * Prints CSS and JavaScript for Leaflet base.
 *
 * @internal
 */
function routes_print_leaflet() {
  global $routesLeafletElev;

  $vendor_dir = plugin_dir_url(__FILE__) . 'vendor/';
  print '<link rel="stylesheet" href="' . $vendor_dir . 'leaflet/leaflet.css" />' . "\n";
  print '<script src="' . $vendor_dir . 'leaflet/leaflet.js"></script>' . "\n";
  if ($routesLeafletElev) {
    print '<script src="' . $vendor_dir . 'd3/d3.min.js"></script>' . "\n";
    print '<link rel="stylesheet" href="' . $vendor_dir . 'leaflet.elevation/leaflet.elevation-0.0.4.css" />' . "\n";
    print '<script type="text/javascript" src="' . $vendor_dir . 'leaflet.elevation/leaflet.elevation-0.0.4.min.js" ></script>' . "\n";
  }
}

/**
 * Prints JavaScript code for the icons.
 *
 * @internal
 */
function routes_print_icons() {
  global $routesIconTable;
  global $wpdb;

?>
<script type="text/javascript">
//<![CDATA[
var routes_icons = [];
<?php
  $qry = "SELECT * FROM $routesIconTable ORDER BY icon_id";
  $icons = $wpdb->get_results($qry);
  foreach ($icons as $icon) {
    print 'routes_icons[' . $icon->icon_id . '] = L.icon({' .
      'iconUrl: \'' . $icon->icon_image . "',\n" .
      'iconSize: [' . $icon->size_x . ', ' . $icon->size_y . "],\n" .
      'iconAnchor: [' . $icon->anchor_x . ', ' . $icon->anchor_y . "],\n";
    if ($icon->shadow_image) {
      print 'shadowUrl: \'' . $icon->shadow_image . "',\n" .
        'shadowSize: [' . $icon->shadow_size_x . ', ' . $icon->shadow_size_y . "],\n" .
        'shadowAnchor: [' . $icon->shadow_anchor_x . ', ' . $icon->shadow_anchor_y . "],\n";
    }
    print 'popupAnchor: [0,0]});' . "\n";
  }
?>
//]]>
</script>
<?php
}

/**
 * Prints the JavaScript header for a non-admin page that will display a map.
 */
function routes_js_header() {
  routes_print_leaflet();
  routes_print_icons();
  wp_enqueue_script('routes-page-js');
}

/**
 * Adds a where clause to post searching, to search by city and state.
 */
function routes_search_where($where) {
  global $wp_query, $wpdb;
  global $routesWaypointTable;

  if (routes_is_search()) {

    // First: searching for general keyword
    // is currently searching on post title and post content.
    // We are going to do what it does, but also add in route_city and
    // route_state.
    // Basically, look for OR (post_title LIKE 'something') and add in
    // our fields too.

    // Try matching on post_title field without table name
    // (old versions of WP)
    $where = preg_replace("/\(\s*post_title\s+LIKE\s*(\'[^\']+\')\s*\)/",
       "(post_title LIKE \\1) OR ($routesWaypointTable.city LIKE \\1) OR ($routesWaypointTable.state_country LIKE \\1) ", $where);

    // Try matching on post_title field with table name
    // (new versions of WP)
    $where = preg_replace("/\(\s*([a-zA-Z0-9_]+)\.post_title\s+LIKE\s*(\'[^\']+\')\s*\)/",
       "(\\1.post_title LIKE \\2) OR ($routesWaypointTable.city LIKE \\2) OR ($routesWaypointTable.state_country LIKE \\2) ", $where);

    // Next: our specific search terms

    $q = $wp_query->query_vars;

    if (!empty($q['routes_city'])) {
      $tmp = stripslashes($q['routes_city']);
      $tmp = $wpdb->escape($tmp);
      $where .= " AND ($routesWaypointTable.city LIKE '%$tmp%') ";
    }

    if (!empty($q['routes_state'])) {
      $tmp = stripslashes($q['routes_state']);
      $tmp = addslashes_gpc($tmp);
      $where .= " AND ($routesWaypointTable.state_country LIKE '%$tmp%') ";
    }
  }

  return $where;
}

/**
 * Adds a join clause to post searching, to search by city and state.
 */
function routes_search_join($join) {
  global $routesWaypointTable, $wpdb, $wp_query;

  if (routes_is_search()) {
    $join .= " LEFT JOIN $routesWaypointTable ON " . $wpdb->posts . ".ID = " .
      "$routesWaypointTable.post_id ";
  }

  return $join;
}

/**
 * Adds GROUP BY clause for searching.
 */
function routes_search_groupby($groupby) {
  global $wpdb, $wp_query;

  if (!routes_is_search()) {
    return $groupby;
  }

  // we need to group on post ID

  $mygroupby = "{$wpdb->posts}.ID";

  if (preg_match("/$mygroupby/", $groupby)) {
    // already there
    return $groupby;
  }

  if (!strlen(trim($groupby))) {
    // was empty, use ours
    return $mygroupby;
  }

  // wasn't empty, append ours
  return $groupby . ", " . $mygroupby;
}

/**
 * Fixes search query vars for our new search terms.
 */
function routes_search_fixquery($notused) {
  global $wp_query;

  if (!routes_is_search()) {
    return;
  }

  $q = $wp_query->query_vars;

  if (!strlen(trim($q['s']))) {
    unset($q['s']);
  }

  if (!strlen(trim($q['routes_city']))) {
    unset($q['routes_city']);
  }

  if (!strlen(trim($q['routes_state']))) {
    unset($q['routes_state']);
  }

  unset($q['routes_addr']);

  $wp_query->query_vars = $q;

}

/**
 * Adds the routes search terms to query vars list.
 */
function routes_search_queryvars($qvars) {
  $qvars[] = 'routes_state';
  $qvars[] = 'routes_city';
  return $qvars;
}

/**
 * Sets is_home to false if we're on a search.
 */
function routes_set_ishome($notused) {
  global $wp_query;

  if (routes_is_search()) {
    $wp_query->is_home = false;
  }
}

/**
 * Adds GPX mime type to allowed types array.
 */
function routes_add_mimes($types) {
  global $routesGPXMimeType;

  $types[ 'gpx' ] = $routesGPXMimeType;
  return $types;

}

/**
 * Prints out a GPX file containing the waypoints, and exits.
 */
function routes_send_gpx($post_id) {
  global $routesGPXMimeType;
  global $routesUseMetric;

  header('Content-type: ' . $routesGPXMimeType);
  header('Content-Disposition: attachment; filename="route.gpx"');

  print '<?xml version="1.0" encoding="ISO-8859-1" standalone="yes"?>';
  print '<gpx version="1.1" creator="Routes WordPress Plugin"><trk><trkseg>' . "\n";

  $post_id = intval($post_id);
  $wpts = array();
  if ($post_id) {
    $wpts = routes_read_waypoints($post_id);
  }
  foreach ($wpts as $waypt) {
    $elev = $waypt->elevation;
    if (!$routesUseMetric) {
      // Elevations in GPX files should be in meters always.
      $elev *= 0.3048;
    }

    print '<trkpt lat="' . $waypt->latitude . '" lon="' . $waypt->longitude . '"><ele>' . $elev . '</ele></trkpt>' . "\n";
  }

  print "</trkseg></trk></gpx>";
  exit();
}

/**
 * Figures out where to put routes graphs.
 *
 * @return array
 *   Array of directory information.
 *
 * @internal
 */
function routes_graph_dir() {
  // code from wp_upload_dir, current and older versions...

  $upload_path = trim(get_option('upload_path'));
  $upload_suff = 'uploads';
  if (defined('UPLOADS')) {
    $upload_suff = UPLOADS;
  }

  if ($upload_path === '') {
    $upload_path = WP_CONTENT_DIR . '/' . $upload_suff;
  }

  $upload_path = str_replace(ABSPATH, '', $upload_path);
  $dir = ABSPATH . $upload_path;

  $url = get_option('upload_url_path');
  if (!$url) {
    $url = trailingslashit(get_option('siteurl')) . $upload_path;
  }

  $dir = $dir . "/routes";
  $url = $url . "/routes";

  if (!wp_mkdir_p($dir)) {
    return array('error' => true);
  }

  return array('path' => $dir, 'url' => $url, 'error' => false);
}

/**
 * Calculates and returns the elevation graph file name.
 *
 * @param int $post_id
 *   ID to make the graph name for.
 *
 * @return string
 *   File name for the elevation graph.
 *
 * @internal
 */
function routes_elev_graph_filename($post_id) {
  return "routes_elev" . $post_id . ".png";
}

/**
 * Replaces the special [[routes_map(L, W, "text")]] tag in a post with map.
 *
 * @internal
 */
function routes_map_in_post($text) {
  global $id;

  $matches = array();

  $pattern = "/\[\[\s*routes_map\s*\(\s*([0-9]+)\s*,\s*([0-9]+)\s*,\s*\"([^\)\"]*)\"\s*\)\s*\]\]/";

  if (!preg_match($pattern, $text, $matches)) {
    return $text;
  }

  $output = '';

  if (is_single() || is_page()) {

    // do what the routes_map function would do, but put it in the text instead
    // only on a single-post page!

    ob_start();
    routes_map($id, $matches[1], $matches[2], $matches[3]);
    $output = ob_get_contents();
    ob_end_clean();

  }

  $text = preg_replace($pattern, $output, $text);

  return $text;
}

/**
 * Returns IMG tags for the given icon's image and shadow.
 *
 * @param int $icon_id
 *   ID of the icon.
 * @param string $alt_img
 *   Alt text for the icon image.
 * @param string $alt_shadow
 *   Alt text for the shadow image.
 *
 * @return array
 *   Array of HTML tags to display the icon and its shadow.
 *
 * @internal
 */
function routes_get_icon_images($icon_id, $alt_img, $alt_shadow) {
  global $wpdb;
  global $routesIconTable;

  // Sanitize.
  $id = intval($icon_id);
  $qry = "SELECT * FROM $routesIconTable WHERE icon_id = $icon_id";
  $stuff = $wpdb->get_results($qry);
  foreach ($stuff as $icon) {
    return array(
      '<img src="' . $icon->icon_image . '" alt="' . $alt_img . '" />',
      '<img src="' . $icon->shadow_image . '" alt="' . $alt_shadow . '" />',
    );
  }
}

/**
 * Prints various types of maps.
 *
 * @param array $wpts
 *   Output of routes_read_waypoints().
 * @param array $poi
 *   Output of routes_read_poi().
 * @param int $width
 *   Width in pixels.
 * @param int $height
 *   Height in pixels
 * @param string type
 *   Type of map:
 *   - 'post': Map suitable for a single post or page.
 *   - 'multiple': Map suitable for home, category page, or search results.
 *   - 'admin': Map suitable for post editing screen.
 * @param bool $newDiv
 *   If TRUE, include printing the div to put the map in.
 * @param string $printNote
 *   Note to print above map (omitted if there is nothing to display).
 *
 * @internal
 */
function routes_map_print($wpts, $poi, $width, $height, $type, $newDiv, $printnote) {
  global $routesBlankLines;
  global $routesJSForFooter;
  global $routesScale;
  global $routesLeafletElev;
  global $routesUseMetric;

  if (!count($wpts) && ($type == "post" || $type == "multiple")) {
    // do not print an empty map for posts
    return;
  }

  echo '<p class="maplegend">' . $printnote . "</p>";

  // Pre-process data

  $routes_as_geo_json = (($type == 'post') && $routesLeafletElev);
  $titles = array();
  $routes = array();
  $orders = array();
  $midpoints = array();
  $aggicons = array();
  $linestyles = array();
  $urls = array();
  $adm_titles = array();
  $tagtext = "";
  $nonvalue = -900000;
  $wpcount = 0;

  if (count($wpts)) {
    $minlat = 90;
    $maxlat = -90;
    $minlong = 180;
    $maxlong = -180;

    $cur_route = array();
    $cur_orders = array();
    $cur_adm_titles = array();
    $cur_id = $wpts[0]->post_id;
    $cur_title = "";
    $cur_icons = routes_get_post_icons($cur_id);
    $cur_style = routes_get_post_line_style($cur_id);

    $cur_mplat = $nonvalue;
    $cur_mplong = $nonvalue;
    $lastlat = $nonvalue;
    $lastlong = $nonvalue;

    foreach ($wpts as $waypoint) {
      // Verify it has a latitude and longitude...
      $this_lat = $waypoint->latitude;
      $this_long = $waypoint->longitude;
      if (!is_numeric($this_lat) || !is_numeric($this_long)) {
        continue;
      }
      if ($cur_mplat == $nonvalue) {
        $cur_mplat = $this_lat;
        $cur_mplong = $this_long;
        $lastlat = $this_lat;
        $lastlong = $this_long;
        $cur_id = $waypoint->post_id;
        $cur_icons = routes_get_post_icons($cur_id);
        $cur_style = routes_get_post_line_style($cur_id);
      }
      $wpcount++;

      if ($waypoint->post_id != $cur_id) {
        // just got to a new post ID, save previous info

        $titles[] = routes_fix_title_js($cur_title);
        $routes[] = $cur_route;
        $orders[] = $cur_orders;
        $aggicons[] = $cur_icons[0];
        $linestyles[] = $cur_style;
        $adm_titles[] = $cur_adm_titles;
        $urls[] = get_permalink($cur_id);

        $cur_mplat = ($cur_mplat + $lastlat) * 0.5;
        $cur_mplong = ($cur_mplong + $lastlong) * 0.5;

        $midpoints[] = "L.latLng(" . routes_commaconvert($cur_mplat) .
          ", " . routes_commaconvert($cur_mplong) . ")";

        // reset for next set

        $cur_route = array();
        $cur_orders = array();
        $cur_adm_titles = array();
        $cur_id = $waypoint->post_id;
        $cur_mplat = $this_lat;
        $cur_mplong = $this_long;
        $cur_icons = routes_get_post_icons($cur_id);
        $cur_style = routes_get_post_line_style($cur_id);
      }

      // store current info
      $latlng = "L.latLng(" . routes_commaconvert($this_lat) . ',' . routes_commaconvert($this_long) . ")";
      if ($routes_as_geo_json) {
        // GeoJSON format is: longitude, latitude, elevation!
        $cur_route[] = '[' . routes_commaconvert($this_long) . ',' . routes_commaconvert($this_lat) . ',' . routes_commaconvert($waypoint->elevation) . ']';
      }
      else {
        $cur_route[] = $latlng;
      }
      $cur_orders[] = $waypoint->order;
      $cur_title = $waypoint->post_title;
      $cur_adm_titles[] = $waypoint->waypoint_order . ": " .
        $this_lat . ", " .
        $this_long;

      $indx = count($cur_route);

      if ($indx == 1 ) {
        // Add the start icon.
        $tagtext .= "routes_add_marker($latlng, '" .
          routes_fix_info_window_js(trim($waypoint->tag)) . "', " .
          $cur_icons[1] . ");\n";
      }
      else if (strlen(trim($waypoint->tag)) > 0) {
        // Add a tag icon.
        $tagtext .= "routes_add_marker($latlng, '" .
          routes_fix_info_window_js(trim($waypoint->tag)) . "', " .
          $cur_icons[2] . ");\n";
      }

      $lastlat = $this_lat;
      $lastlong = $this_long;

      $maxlat = max($maxlat, $this_lat);
      $maxlong = max($maxlong, $this_long);
      $minlat = min($minlat, $this_lat);
      $minlong = min($minlong, $this_long);

    } // end of loop over waypoints

    // add remaining info from last route
    $titles[] = routes_fix_title_js($cur_title);
    $routes[] = $cur_route;
    $orders[] = $cur_orders;
    $aggicons[] = $cur_icons[0];
    $linestyles[] = $cur_style;
    $adm_titles[] = $cur_adm_titles;
    $urls[] = get_permalink($cur_id);

    $cur_mplat = ($cur_mplat + $lastlat) * 0.5;
    $cur_mplong = ($cur_mplong + $lastlong) * 0.5;
    $midpoints[] = "L.latLng(" . routes_commaconvert($cur_mplat) .
      ", " . routes_commaconvert($cur_mplong) . ")";

  }
  if (!$wpcount) {
    // set up for whole world map, as we have no waypoints
    $minlat = -90;
    $maxlat = 90;
    $minlong = -180;
    $maxlong = 180;
  }

  $midlat = ($minlat + $maxlat) * 0.5;
  $midlong = ($minlong + $maxlong) * 0.5;

  // Print out div and other stuff (depending on type of map)

  if ($type == "admin") {

    echo "<p>";
    echo '<input type="button" name="routes_refresh_btn" id="routes_refresh_btn" value="' .
    __("Refresh Map", 'routes') . '" ' .
    ' onclick="routes_refresh_map();"' .
    "/> ";
    echo '<input type="radio" name="routes_click_action" value="waypoint" checked="checked" /> ';
    _e('Clicking on map adds route point', 'routes');
    echo '<input type="radio" name="routes_click_action" value="poi" /> ';
    _e('Clicking on map adds point of interest', 'routes');
    echo "<br />\n";
    _e("Note: If you have edited Order fields above, click Save and Continue Editing to refresh map.", 'routes');
    echo " ";
    _e("Click on a marker to remove a waypoint or point of interest. Click on a new location to add a waypoint or point of interest, then enter or edit other fields above.", 'routes');
    echo " ";
    _e("Note: If you have a lot of route points, only about 50 will be shown as markers in the map, to save loading time, but all will be shown on the route.", 'routes');

    echo "</p>";
  }

  if ($newDiv) {
    echo "<div id=\"routes_map\" style=\"width: $width" .
      "px; height: $height" . "px\"></div><br />";
  }

  // Create the map with attributions, and make it show all current
  // latitudes/longitudes.
  // Actually we'll save this for later, because IE doesn't work to have
  // JavaScript in a div, for some reason.

  $routesJSForFooter .= "var routesmap;\n";
  $routesJSForFooter .= "routesmap = L.map('routes_map');\n";
  $routesJSForFooter .= "var osmUrl='https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png';\n";
  $routesJSForFooter .= "var osmAttrib='Map data © <a href=\"https://openstreetmap.org\">OpenStreetMap</a> contributors';\n";
  $routesJSForFooter .= "var osm = L.tileLayer(osmUrl, {attribution: osmAttrib});\n";

  // center and zoom map
  // If using the Leaflet plugin, add some padding on the top.
  if ($routes_as_geo_json) {
    $frac = 100 / $height;
    $latdiff = $maxlat - $minlat;
    $maxlat += $frac * $latdiff;
  }

  $midlat = routes_commaconvert($midlat);
  $midlong = routes_commaconvert($midlong);
  $minlat = routes_commaconvert($minlat);
  $minlong = routes_commaconvert($minlong);
  $maxlat = routes_commaconvert($maxlat);
  $maxlong = routes_commaconvert($maxlong);

  if ($wpcount > 1) {
    $routesJSForFooter .= "routesmap.fitBounds(" .
      " L.latLngBounds(" .
      "   L.latLng($minlat, $minlong), " .
      "   L.latLng($maxlat, $maxlong)));\n";

  }
  else {
    $routesJSForFooter .= "routesmap.setView(" .
      "L.latLng($midlat, $midlong), 10);\n";
  }
  $routesJSForFooter .= "routesmap.addLayer(osm);\n";

  // Add route(s)

  if ($type == "post") {

    for ($i = 0; $i < count($titles); $i++) {
      if (count($routes[ $i ]) > 1) {
        // add route line to map, if it has at least two points
        if ($routes_as_geo_json) {
          $routesJSForFooter .= "var el = L.control.elevation({\n";
          if (!$routesUseMetric) {
            $routesJSForFooter .=
              "  imperial: true,\n";
          }
          $elev_width = $width - 100;
          $routesJSForFooter .=
            "  width: " . $elev_width . ",\n" .
            "  height: 100,\n" .
            "  useHeightIndicator: false,\n" .
            "});\n";
          $routesJSForFooter .= "var geo_json = {\n" .
            "type: 'FeatureCollection',\n" .
            "features:[{\n" .
            "  type: 'Feature',\n" .
            "  geometry: {\n" .
            "    type: 'LineString',\n" .
            "    coordinates: [\n" .
            "      " . implode(",\n      ", $routes[$i]) . "\n" .
            "    ]\n" .
            "  }\n" .
            "}]};\n";
          $routesJSForFooter .= "var lgs = L.geoJson(geo_json, {\n" .
            "  onEachFeature: el.addData.bind(el)\n" .
            "});\n";
        }
        else {
          $routesJSForFooter .= "var routesline = L.polyline(new Array(";
          $routesJSForFooter .= implode(", ", $routes[ $i ]);
          $routesJSForFooter .= '), { color: "#' . $linestyles[$i][0] . '", weight: ' . $linestyles[$i][2] . ', opacity: ' . routes_commaconvert($linestyles[$i][1]) . "});\n";
          $routesJSForFooter .= "routesline.addTo(routesmap);\n";
        }
      }
    }

    // add markers for tagged points
    $routesJSForFooter .= $tagtext;

    // Add POI, if there are any.
    if (count($poi)) {
      foreach ($poi as $point) {
        $this_lat = $point->latitude;
        $this_long = $point->longitude;
        if (!is_numeric($this_lat) || !is_numeric($this_long)) {
          continue;
        }

        $latlng = "L.latLng(" . routes_commaconvert($this_lat) . ',' . routes_commaconvert($this_long) . ")";
        $indx++;

        $routesJSForFooter .= "routes_add_marker($latlng, '" .
          routes_fix_info_window_js(trim($point->tag)) . "', " .
          $point->icon_id . ");\n";
      }
    }

    // If this is a GeoJSON map, now add the elevation and route so they
    // are on top.
    if ($routes_as_geo_json) {
      $routesJSForFooter .=
        "el.addTo(routesmap);\n" .
        "lgs.addTo(routesmap);\n";
    }
  } else if ($type == "multiple") {

    $routesJSForFooter .= "var routes_mrkr;\n";

      // create small icon

    for ($i = 0; $i < count($titles); $i++) {
      if (count($routes[ $i ]) > 1) {
        // add straight route line to map, if it has at least two points,
        // with event listener that opens URL if line is clicked
        $routesJSForFooter .= "routes_rtln = L.polyline(new Array(";
        $routesJSForFooter .= $routes[$i][0] . ", " . $routes[$i][ count($routes[$i]) - 1] ;
        $routesJSForFooter .= '), {color: "#' . $linestyles[$i][0] . '", weight: ' . $linestyles[$i][2] . ', opacity: ' . routes_commaconvert($linestyles[$i][1]) . ", clickable:true});\n";

        $routesJSForFooter .=  "routes_rtln.on(\"click\", function() { " .
        "location=\"" . $urls[$i] . "\"});\n";
        $routesJSForFooter .= "routes_rtln.addTo(routesmap);\n";
      }

      // add midpoint of the route as a marker, with event listener
      // that opens URL if the marker is clicked

      $routesJSForFooter .=  "routes_mrkr = L.marker(" . $midpoints[$i] .
        ", { title: '" . $titles[$i] . "', icon: routes_icons[" . $aggicons[$i] . "]}).on('click', function() { location=\"" . $urls[$i] . "\"; }).addTo(routesmap);";

    }
  } else { // admin type

    // add num lines variable
    $numlines = count($routes[0]) + $routesBlankLines;
    $numpoi = count($poi) + $routesBlankLines;
    $routesJSForFooter .= "var routesOverlays = new Array();\n";
    $routesJSForFooter .= "var routesNumOverlays = 0;\n";

    $routesJSForFooter .= "var routesFormNumLines = $numlines;\n";
    $routesJSForFooter .= "var routesFormNumPOILines = $numpoi;\n";
    $routesJSForFooter .= "var routesInfoHidden = 1;\n";

    // add detection of the current form
    $routesJSForFooter .=  'function routes_form_ptr() { return document.getElementById(\'routes_refresh_btn\').form; }' . "\n";

    // use the refresh map function to add the points
    $routesJSForFooter .= "routes_refresh_map();\n";

    // add event handler to admin map so click adds a new point.
    $routesJSForFooter .= 'routesmap.on("click", function(event) {' . "\n";
    $routesJSForFooter .= "point = event.latlng;\n";
    $routesJSForFooter .= "routes_add_point(point.lat, point.lng);\n";
    $routesJSForFooter .= "\n";
    $routesJSForFooter .= "routes_refresh_map();\n";
    $routesJSForFooter .= "});\n";
  } // end of if/else on type of map (post/multiple/admin)

}

/**
 * Makes sure proper setup has been done.
 *
 * Reads options and sets up internationalization.
 */
function routes_setup($force = 0) {
  global $routesIsSetUp;
  global $routesScale;
  global $routesUseMetric;
  global $routesUseStartIcon;
  global $routesMakeElev;
  global $routesLookupElev;
  global $routesLeafletElev;
  global $routesElevMarkers;
  global $routesElevWidth, $routesElevHeight;
  global $routesElevFillColor;
  global $routesElevBgColor;
  global $routesElevLineColor;
  global $routesElevMarkerColor;
  global $routesWidgetSearchTitle, $routesWidgetSearchKeywordLabel,
    $routesWidgetSearchCityLabel,
    $routesWidgetSearchStateLabel,
    $routesWidgetSearchSearchLabel;
  global $routesWidgetStatesTitle, $routesWidgetStatesSep;

  if ($routesIsSetUp && !$force) {
    return;
  }

  // set up internationalization

  load_plugin_textdomain('routes',
                          'wp-content/plugins', '');

  $routesUseMetric = routes_get_yn_option('routes_use_metric');
  $routesUseStartIcon = routes_get_yn_option('routes_use_start_icon');
  $routesMakeElev = routes_get_yn_option('routes_make_elev');
  $routesLookupElev = routes_get_yn_option('routes_lookup_elev');
  $routesLeafletElev = routes_get_yn_option('routes_leaflet_elev');
  $routesElevMarkers = routes_get_yn_option('routes_elev_markers');

  $routesElevWidth = get_option('routes_elev_width');
  if ($routesElevWidth <= 0) {
    $routesElevWidth = 500;
  }

  $routesElevHeight = get_option('routes_elev_height');
  if ($routesElevHeight <= 0) {
    $routesElevHeight = 100;
  }

  $routesElevFillVal = get_option('routes_elev_fill_color');
  if ($routesElevFillVal) {
    $routesElevFillColor = explode(",", $routesElevFillVal);
  }

  $routesElevBgVal = get_option('routes_elev_bg_color');
  if ($routesElevBgVal) {
    $routesElevBgColor = explode(",", $routesElevBgVal);
  }

  $routesElevLineVal = get_option('routes_elev_line_color');
  if ($routesElevLineVal) {
    $routesElevLineColor = explode(",", $routesElevLineVal);
  }

  $routesElevMarkerVal = get_option('routes_elev_marker_color');
  if ($routesElevMarkerVal) {
    $routesElevMarkerColor = explode(",", $routesElevMarkerVal);
  }

  $routesWidgetSearchKeywordLabel = get_option('routes_widget_search_keyword_label');
  if ($routesWidgetSearchKeywordLabel === false) {
    $routesWidgetSearchKeywordLabel = __('Keywords', 'routes');
  }

  $routesWidgetSearchTitle = get_option('routes_widget_search_title');
  if ($routesWidgetSearchTitle === false) {
    $routesWidgetSearchTitle = __('Search', 'routes');
  }

  $routesWidgetSearchCityLabel = get_option('routes_widget_search_city_label');
  if ($routesWidgetSearchCityLabel === false) {
    $routesWidgetSearchCityLabel = __('City', 'routes');
  }

  $routesWidgetSearchStateLabel = get_option('routes_widget_search_state_label');
  if ($routesWidgetSearchStateLabel === false) {
    $routesWidgetSearchStateLabel = __('State', 'routes');
  }

  $routesWidgetSearchSearchLabel = get_option('routes_widget_search_search_label');
  if ($routesWidgetSearchSearchLabel === false) {
    $routesWidgetSearchSearchLabel = __('Search', 'routes');
  }

  $routesWidgetStatesTitle = get_option('routes_widget_states_title');
  if ($routesWidgetStatesTitle === false) {
    $routesWidgetStatesTitle = __('States', 'routes');
  }

  $routesWidgetStatesSep = get_option('routes_widget_states_sep');
  if ($routesWidgetStatesSep === false) {
    $routesWidgetStatesSep = ' | ';
  }

  // no need to do this stuff more than once

  $routesIsSetUp = 1;
}

/**
 * Prints out JavaScript to set up the map.
 *
 * Must be put after all the DIVs are closed at the bottom of the page footer,
 * to work in all browsers.
 *
 * @internal
 */
function routes_do_map_js() {
  global $routesJSForFooter;
  if ($routesJSForFooter) {
    echo '<script type="text/javascript">' . "\n";
    echo "//<![CDATA[\n";

    echo $routesJSForFooter;

    echo "//]]>\n";
    echo "</script>\n";
  }
}

/**
 * Reads waypoint data from the database for a route or list of routes.
 *
 * @param int|array $route_list
 *   ID of post or page to read for, or an array of IDs.
 *
 * @return array
 *   Array of database rows with waypoint data, with addition of post_title
 *   field. Data is ordered by route ID first, and within route ID, by waypoint
 *   order.
 *
 * @internal
 */
function routes_read_waypoints($route_list) {
  global $wpdb, $routesWaypointTable;

  if (!$route_list) {
    $route_list = array(0);
  }

  // convert route list into a WHERE clause

  if (!is_array($route_list)) {
    $route_list = array($route_list);
  }

  $wheres = array();
  foreach ($route_list as $route) {
    $wheres[] = "(post_id = '" . $wpdb->escape($route) . " ')";
  }

  // build query, run, and return results

  $query = "SELECT waypoint_id, post_id, waypoint_order, " .
    "city, state_country, latitude, longitude, elevation, tag, post_title " .
    " FROM $routesWaypointTable LEFT JOIN " . $wpdb->posts . " ON  " .
    $routesWaypointTable . ".post_id = " . $wpdb->posts . ".ID " .
    " WHERE " .
    implode(" OR ", $wheres) . " ORDER BY post_id, waypoint_order";

  return $wpdb->get_results($query);
}

/**
 * Reads points of interest data from the database for a route.
 *
 * @param int $post_id
 *   ID of post or page to read for.
 *
 * @return array
 *   Array of database rows with POI data.
 *
 * @internal
 */
function routes_read_poi($post_id) {
  global $wpdb, $routesPOITable;

  $post_id = intval($post_id);
  $query = "SELECT * from $routesPOITable WHERE post_id = " . $post_id;
  return $wpdb->get_results($query);
}

/**
 * Fixes up post titles for use in JavaScript.
 *
 * @param string $title
 *   Post or page title.
 *
 * @return string
 *   Fixed-up title.
 *
 * @internal
 */
function routes_fix_title_js($title) {
  $text = strip_tags(apply_filters("the_title", $title));

  $text = ent2ncr($text);

  $ncr_list = array(
                    '&#034;' => "\'",
                    '&#8217;' => "\'",
                    '&#038;' => '&',
                    '&#047;' => '/',
                    '&#0124;' => '|'
       );

  $text = str_replace(array_keys($ncr_list), array_values($ncr_list), $text);

  // any remaining numerical entities, let JavaScript fix

  $text = preg_replace('/&#(\d+);/', '\' + String.fromCharCode(\\1) + \'', $text);

  return $text;
}

/**
 * Fixes text for display in JavaScript.
 *
 * @param string $text
 *   Text to fix.
 *
 * @return string
 *   Text with certain characters escaped.
 *
 * @internal
 */
function routes_fix_info_window_js($text) {
  // fix HTML entities

  $text = ent2ncr($text);

  $ncr_list = array(
                    '&#034;' => "'",
                    '&#8217;' => "'",
                    '&#038;' => '&',
                    '&#047;' => '/',
                    '&#0124;' => '|'
       );

  $text = str_replace(array_keys($ncr_list), array_values($ncr_list), $text);

  // any remaining numerical entities, let JavaScript fix

  $text = preg_replace('/&#(\d+);/', '\' + String.fromCharCode(\\1) + \'', $text);

  $text = preg_replace("|'|", '&rsquo;', stripslashes($text));

  return $text;
}

/**
 * Converts , to . in numbers, in case of European formats, for JS.
 *
 * @param string $numtext
 *   Textual number.
 *
 * @return string
 *   String number with commas converted to . so it can be read as a number
 *   by JavaScript.
 *
 * @internal
 */
function routes_commaconvert($numtext) {
  return str_replace(",", ".", $numtext);
}

/**
 * Gets a Y/N option and coverts to 1/0.
 *
 * @param string $optname
 *   Option name to convert, or text to convert.
 * @param int $as_txt
 *   If 0, assumes $optname is the name of the option. If 1, assumes it is the
 *   actual Y/N text.
 *
 * @return int
 *   Converted option.
 *
 * @internal
 *
 * @see routes_num_to_yn
 */
function routes_get_yn_option($optname, $as_txt = 0) {
  $txt = $optname;
  if (!$as_txt) {
    $txt = get_option($optname);
  }

  if ($txt == 'y' || $txt == 'Y') {
     return 1;
  }

  return 0;
}

/**
 * Constrains a number to lie between min and max, and returns it.
 *
 * @internal
 */
function routes_fix_num($num, $min, $max) {
  if ($num > $max) {
    $num = $max;
  }

  if ($num < $min) {
    $num =  $min;
  }

  return $num;
}
